package com.ultraboodog.level;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.imageio.ImageIO;

import com.ultraboodog.entity.Entity;
import com.ultraboodog.entity.mob.Player;
import com.ultraboodog.entity.particle.Particle;
import com.ultraboodog.graphics.Screen;
import com.ultraboodog.level.tile.Tile;
import com.ultraboodog.util.Vector2i;

public class Level
{
	public int level = 1;
	
	protected int width, height;
	protected int[] tiles;

	private List<Entity> entities = new ArrayList<Entity>();
	private List<Particle> particles = new ArrayList<Particle>();
	private List<Player> players = new ArrayList<Player>();
	
	private Comparator<Node> nodeSorter = new Comparator<Node>()
	{
		public int compare(Node n0, Node n1)
		{
			if (n1.fCost < n0.fCost) return 1;
			if (n1.fCost > n0.fCost) return -1;
			return 0;
		}
	};
	
	public Level(String path)
	{
		loadLevel(path);
	}
	
	protected void generateLevel()
	{
	}
	
	protected void loadLevel(String path)
	{
		try
		{
			BufferedImage image = ImageIO.read(WaterLevel1.class.getResource(path));
			int w = width = image.getWidth();
			int h = height = image.getHeight();
			tiles = new int[w * h];
			image.getRGB(0, 0, w, h, tiles, 0, w);
		} catch (IOException e)
		{
			e.printStackTrace();
			System.err.println("Could not load level file!");
		}
	}
	
	public void update()
	{
		for (int i = 0; i < entities.size(); i++)
			entities.get(i).update();
		
		for (int i = 0; i < particles.size(); i++)
			particles.get(i).update();
		
		for (int i = 0; i < players.size(); i++)
			players.get(i).update();
		
		remove();
	}
	
	private void remove()
	{
		for (int i = 0; i < entities.size(); i++)
			if (entities.get(i).isRemoved()) entities.remove(i);
		
		for (int i = 0; i < particles.size(); i++)
			if (particles.get(i).isRemoved()) particles.remove(i);
		
		for (int i = 0; i < players.size(); i++)
			if (players.get(i).isRemoved()) players.remove(i);
	}
	
	public boolean tileCollision(int x, int y, int size, int xOffset, int yOffset)
	{
		boolean solid = false;
		for (int c = 0; c < 4; c++)
		{
			int xt = (x - c % 2 * size + xOffset) >> 4;
			int yt = (y - c / 2 * size + yOffset) >> 4;
			if (getTile(xt, yt).solid()) solid = true;
		}
		return solid;
	}
	
	public void render(int xScroll, int yScroll, Screen screen)
	{
		// Sets the map boundaries
		if (xScroll < 16) xScroll = 16;
		if (xScroll > (width << 4) - screen.width) xScroll = (width << 4) - (screen.width);
		if (yScroll < 16) yScroll = 16;
		if (yScroll > (height << 4) - screen.height) yScroll = (height << 4) - (screen.height);
		
		screen.setOffset(xScroll, yScroll);
		// Corner Pins
		int x0 = xScroll >> 4;
		int x1 = (xScroll + screen.width + 16) >> 4;
		int y0 = yScroll >> 4;
		int y1 = (yScroll + screen.height + 16) >> 4;
		
		// Render
		for (int y = y0; y < y1; y++)
		{
			for (int x = x0; x < x1; x++)
			{
				getTile(x, y).render(x, y, screen);
			}
		}
		
		for (int i = 0; i < entities.size(); i++)
			entities.get(i).render(screen);
		
		for (int i = 0; i < particles.size(); i++)
			particles.get(i).render(screen);
		
		for (int i = 0; i < players.size(); i++)
			players.get(i).render(screen);
	}
	
	public void add(Entity e)
	{
		e.init(this);
		if (e instanceof Particle) particles.add((Particle) e);
		else if (e instanceof Player) players.add((Player) e);
		else entities.add(e);
	}
	
	public List<Node> findPath(Vector2i start, Vector2i goal)
	{
		List<Node> openList = new ArrayList<Node>();
		List<Node> closedList = new ArrayList<Node>();
		Node current = new Node(start, null, 0, getDistance(start, goal));
		openList.add(current);
		while (openList.size() > 0)
		{
			Collections.sort(openList, nodeSorter);
			current = openList.get(0);
			if (current.tile.equals(goal))
			{
				List<Node> path = new ArrayList<Node>();
				while(current.parent != null)
				{
					path.add(current);
					current = current.parent;
				}
				openList.clear();
				closedList.clear();
				return path;
			}
			openList.remove(current);
			closedList.add(current);
			for (int i = 0; i < 9; i++)
			{
				if (i == 4) continue;
				int x = current.tile.getX();
				int y = current.tile.getY();
				int xi = (i % 3) - 1;
				int yi = (i / 3) - 1;
				Tile at = getTile(x + xi, y + yi);
				if (at == null) continue;
				if (at.solid()) continue;
				Vector2i a = new Vector2i(x + xi, y + yi);
				double gCost = current.gCost + (getDistance(current.tile, a) == 1 ? 1 : 0.95);
				double hCost = getDistance(a, goal);
				Node node = new Node(a, current, gCost, hCost);
				if (vecInList(closedList, a) && gCost >= node.gCost) continue;
				if (!vecInList(openList, a) || gCost < node.gCost) openList.add(node);
			}
		}
		closedList.clear();
		return null;
	}
	
	private boolean vecInList(List<Node> list, Vector2i vector)
	{
		for (Node n : list)
		{
			if (n.tile.equals(vector)) return true;
		}
		return false;
	}
	
	private double getDistance(Vector2i tile, Vector2i goal)
	{
		double dx = tile.getX() - goal.getX();
		double dy = tile.getY() - goal.getY();
		return Math.sqrt(dx * dx + dy * dy);
	}
	
	public List<Entity> getEntities(Entity e, int radius)
	{
		List<Entity> result = new ArrayList<Entity>();
		int ex = (int) e.getX();
		int ey = (int) e.getY();
		
		for (int i = 0; i < entities.size(); i++)
		{
			Entity entity = entities.get(i);
			int x = (int) entity.getX();
			int y = (int) entity.getY();
			int dx = Math.abs(x - ex);
			int dy = Math.abs(y- ey);
			double distance = Math.sqrt((dx * dx) + (dy * dy));
			
			if (distance <= radius) result.add(entity);
		}
		return result;
	}
	
	public List<Player> getPlayers()
	{
		return players;
	}
	
	public Player getPlayerAt(int index)
	{
		return players.get(index);
	}
	
	public Player getClientPlayer()
	{
		return players.get(0);
	}
	
	public List<Player> getPlayers(Entity e, int radius)
	{
		List<Player> result = new ArrayList<Player>();
		int ex = (int) e.getX();
		int ey = (int) e.getY();
		
		for (int i = 0; i < players.size(); i++)
		{
			Player player = players.get(i);
			int x = (int) player.getX();
			int y = (int) player.getY();
			int dx = Math.abs(x - ex);
			int dy = Math.abs(y - ey);
			double distance = Math.sqrt((dx * dx) + (dy * dy));
			
			if (distance <= radius) result.add(player);
		}
		return result;
	}
	
	public Tile getTile(int x, int y)
	{
		if (x < 0 || y < 0 || x >= width || y >= height) return Tile.voidTile;
		if (tiles[x + y * width] == 0xFF00710B) return Tile.grass;
		if (tiles[x + y * width] == 0xFF7F7F7F) return Tile.stone;
		if (tiles[x + y * width] == 0xFFF9E586) return Tile.sand;
		if (tiles[x + y * width] == 0xFFFF65C5) return Tile.sandCoral;
		if (tiles[x + y * width] == 0xFFFF8533) return Tile.orangePlant;
		if (tiles[x + y * width] == 0xFF6E6640) return Tile.sinkHoleExit;
		if (tiles[x + y * width] == 0xFFD06767) return Tile.shell;
		if (tiles[x + y * width] == 0xFF757249) return Tile.slowSand;
		if (tiles[x + y * width] == 0xFF0371BD) return Tile.blueShell;
		if (tiles[x + y * width] == 0xFFA7A67B) return Tile.fastSand;
		if (tiles[x + y * width] == 0xFF76B13E) return Tile.teleportSand;
		if (tiles[x + y * width] == 0xFF4394B7) return Tile.sandWater;
		if (tiles[x + y * width] == 0xFFFFF596) return Tile.brightSand;
		return Tile.voidTile;
	}
}
